@@ -49,7 +49,7 @@ func TestUnresponsiveConnection(t *testing.T) {
 	if !ok {
 		t.Fatalf("no PacketResponse in response channel")
 	}
-	packet, err = packetResponse.ReadPacket()
+	_, err = packetResponse.ReadPacket()
 	if err == nil {
 		t.Fatalf("expected timeout error")
 	}
@@ -58,6 +58,39 @@ func TestUnresponsiveConnection(t *testing.T) {
 	}
 }
 
+// TestInvalidStateCloseDeadlock tests that we do not enter deadlock when the
+// message handler is blocked or inactive.
+func TestInvalidStateCloseDeadlock(t *testing.T) {
+	// The do-nothing server that accepts requests and does nothing
+	ts := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+	}))
+	defer ts.Close()
+	c, err := net.Dial(ts.Listener.Addr().Network(), ts.Listener.Addr().String())
+	if err != nil {
+		t.Fatalf("error connecting to localhost tcp: %v", err)
+	}
+
+	// Create an Ldap connection
+	conn := NewConn(c, false)
+	conn.SetTimeout(time.Millisecond)
+
+	// Attempt to close the connection when the message handler is
+	// blocked or inactive
+	conn.Close()
+}
+
+// TestInvalidStateSendResponseDeadlock tests that we do not enter deadlock when the
+// message handler is blocked or inactive.
+func TestInvalidStateSendResponseDeadlock(t *testing.T) {
+	// Attempt to send a response packet when the message handler is blocked or inactive
+	msgCtx := &messageContext{
+		id:        0,
+		done:      make(chan struct{}),
+		responses: make(chan *PacketResponse),
+	}
+	msgCtx.sendResponse(&PacketResponse{}, time.Millisecond)
+}
+
 // TestFinishMessage tests that we do not enter deadlock when a goroutine makes
 // a request but does not handle all responses from the server.
 func TestFinishMessage(t *testing.T) {
